/*
 * rtos.c
 *
 * Created: 26.09.2013 10:06:34
 *  Author: Orion
 */ 

#include "rtos.h"

/************************************************************************/
/* RTOS Init                                                            */
/************************************************************************/

inline void RunRTOS()
{
    TIMSK0  |= 1<<TOIE0;         //  System timer
    TCCR0A  |= 0<<WGM01;         //  Compare timer mode, non-PWM
    TCCR0B  |= 0<<CS02
            |  1<<CS01
            |  0<<CS00;          //  start timer, divider 256
}

inline void InitRTOS()
{
    for (uint8_t i = 0; i < TSK_QUE_SIZE; i++)
    {
        TaskQueue[i] = Idle;
    }
    
    for (uint8_t i = 0; i < MAIN_TMR_SIZE; i++)
    {
        MainTimer[i].goToTask = Idle;
        MainTimer[i].delay    = 0;
        
        TimerQueue[i] = &MainTimer[i];  //      
                                        //         
    }
    add_tsk_idx = get_tsk_idx = TSK_QUE_SIZE - 1;
    add_tmr_idx = get_tmr_idx = MAIN_TMR_SIZE -1;
}

/************************************************************************/
/* RTOS Management                                                      */
/************************************************************************/

void Idle(void)
{
    
}

//     .

void SetTimerTask(tsk_ptr_t _task, uint16_t _delay)
{
    cli();
    
    //      FIFO,   .
    //  ..     ,     !
    if (TimerQueue[add_tmr_idx] != 0)     //       , 
    {                                     //  !
        TimerQueue[add_tmr_idx]->goToTask = _task;
        TimerQueue[add_tmr_idx]->delay    = _delay;
        TimerQueue[add_tmr_idx]           = 0;
        //       ,    .
        //      . . ,   
        //    .
        add_tmr_idx = (add_tmr_idx)?            //     0,     
                      add_tmr_idx-1:            //  , 
                      MAIN_TMR_SIZE-1;          //     
    }
    sei();  //      ,    . .
}

//   

void TimerMgr()
{
    for (uint8_t i = 0; i < MAIN_TMR_SIZE; i++)
    {
        if (MainTimer[i].goToTask == Idle) continue;    //    ,  
        if (MainTimer[i].delay != 0)                    //     
        {
            MainTimer[i].delay--;                       //   
        }
        else
        {
            AddTask(MainTimer[i].goToTask);             //      
            MainTimer[i].goToTask = Idle;               //     
            TimerQueue[get_tmr_idx] = &MainTimer[i];    //      
            get_tmr_idx = (get_tmr_idx)?                //     
                          get_tmr_idx-1:                //     ,
                          MAIN_TMR_SIZE-1;              //     add  get 
        }
    }
}

//     .  

void AddTask(tsk_ptr_t _task)
{
    cli();
    //       FIFO,   100% , 
    //     ,    .
    if (TaskQueue[add_tsk_idx] == Idle)     //        
    {                                       //  !
        TaskQueue[add_tsk_idx] = _task;
        //       ,     .
        //      . . ,   
        //     .
        add_tsk_idx = (add_tsk_idx)?        //     0,     
                       add_tsk_idx-1:       //  , 
                       TSK_QUE_SIZE-1;      //     
    }
    sei();                                  //     ,    . .    
}

//    

void TaskMgr()
{
    if (TaskQueue[get_tsk_idx] == Idle) return;     //      , 
    else                                            //   
    {
        cli();      //  ,   ,    
                    //        ?
    
        //      FIFO,       .
        //   ,          ,  .
    
        tsk_ptr_t goToTask = TaskQueue[get_tsk_idx];    //       
        TaskQueue[get_tsk_idx] = Idle;                  //    ,    
        get_tsk_idx = (get_tsk_idx)?                    //     0,     
                       get_tsk_idx-1:                   //  , 
                       TSK_QUE_SIZE-1;                  //     
        //  !       ! (...    )
        sei();
        (goToTask)();                               //   
    }
}

/************************************************************************/
/* System Timer INT                                                     */
/************************************************************************/

ISR(TIMER0_OVF_vect)
{
    TimerMgr();
}